from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import collections
import base64
import datetime

import dateutil.parser
import six

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        import xml.etree.ElementTree as etree


__all__ = ["Serializer"]


class Serializer(object):

    def serialize(self, data):
        if len(data) != 1:
            raise ValueError("Cannot serialize a multi item or empty dictionary")

        if "methodCall" in data:
            xml = etree.Element("methodCall")

            methodName = etree.Element("methodName")
            methodName.text = data["methodCall"]["methodName"]

            xml.append(methodName)

            data = data["methodCall"]
        else:
            raise ValueError("xmlrpc packet not able to be serialized")

        if "params" in data:
            params = etree.Element("params")

            for item in data["params"]:
                param = etree.Element("param")
                param.append(self.to_xml(item))
                params.append(param)

            xml.append(params)

        return etree.tostring(xml)

    def deserialize(self, data):
        xml = etree.fromstring(data)

        if xml.tag == "methodCall":
            python = {
                "methodCall": {},
            }

            methodName = xml.findtext("methodName")

            python["methodCall"]["methodName"] = methodName

            data = python["methodCall"]
        elif xml.tag == "methodResponse":
            python = {
                "methodResponse": {},
            }

            data = python["methodResponse"]
        else:
            raise ValueError("xmlrpc packet not able to be deserialized")

        params = xml.find("params")

        if params is not None:
            data["params"] = []
            for item in params:
                if item.tag == "param":
                    data["params"].append(self.from_xml(item.find("value")))
                else:
                    raise ValueError("Incorrect tag inside of params")

        fault = xml.find("fault")

        if fault is not None:
            data["fault"] = self.from_xml(fault.find("value"))

        return python

    def to_xml(self, data):
        value = etree.Element("value")

        if isinstance(data, six.text_type):
            item = etree.Element("string")
            item.text = data
        elif isinstance(data, bool):
            item = etree.Element("boolean")
            item.text = six.text_type(1 if data else 0)
        elif isinstance(data, int):
            item = etree.Element("int")
            item.text = six.text_type(data)
        elif isinstance(data, float):
            item = etree.Element("double")
            item.text = six.text_type(data)
        elif isinstance(data, datetime.datetime):
            item = etree.Element("dateTime.iso8601")
            item.text = data.isoformat()
        elif isinstance(data, six.binary_type):
            item = etree.Element("base64")
            item.text = base64.b64encode(data)
        elif isinstance(data, collections.Mapping):
            item = etree.Element("struct")

            for k, v in data.items():
                member = etree.Element("member")

                name = etree.Element("name")
                name.text = k

                member.append(name)
                member.append(self.to_xml(v))

                item.append(member)
        elif isinstance(data, collections.Iterable):
            item = etree.Element("array")
            array_data = etree.Element("data")

            for x in data:
                array_data.append(self.to_xml(x))

            item.append(array_data)
        else:
            raise ValueError("Unable to serialize {cls} objects, unknown type".format(cls=data.__class__.__name__))

        value.append(item)

        return value

    def from_xml(self, data):
        if not data.tag == "value":
            raise ValueError("Cannot deserialize a non value")

        # Values without a type are considered strings
        if data.text:
            return data.text

        # XMLRPC packets always have a single item inside of a value
        value = data[0]

        if value.tag == "string":
            return six.text_type(value.text)
        elif value.tag == "boolean":
            return bool(int(value.text))
        elif value.tag in ("int", "i4"):
            return int(value.text)
        elif value.tag == "double":
            return float(value.text)
        elif value.tag == "dateTime.iso8601":
            return dateutil.parser.parse(value.text)
        elif value.tag == "base64":
            return base64.b64decode(value.text.encode("utf-8"))
        elif value.tag == "struct":
            mapping = {}

            for member in value:
                key, value = None, None
                for i in member:
                    if i.tag == "name":
                        key = i.text
                    elif i.tag == "value":
                        value = self.from_xml(i)
                    else:
                        raise ValueError("Unknown struct members")
                mapping[key] = value

            return mapping
        elif value.tag == "array":
            array_data = value.find("data")
            return [self.from_xml(x) for x in array_data]
        else:
            ValueError("Unable to deserialize {type}, unknown type".format(type=value.tag))



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import functools

try:
    import urllib.parse as urllib_parse
except ImportError:
    import urlparse as urllib_parse

from . import requests
from .exceptions import Fault
from .serializer import Serializer


__all__ = ["BaseTransport", "HTTPTransport", "HTTPSTransport", "Client"]


class BaseTransport(object):

    @property
    def scheme(self):
        raise NotImplementedError("Transports must have a scheme")

    def request(self, uri, body):
        raise NotImplementedError("Transports must implement request")


class HTTPTransport(BaseTransport):

    scheme = "http"

    def __init__(self, session=None):
        if session is None:
            session = requests.session()

        self.session = session

    def request(self, uri, body):
        resp = self.session.post(uri, body, headers={"Content-Type": "text/xml"})
        resp.raise_for_status()
        return resp.content


class HTTPSTransport(HTTPTransport):

    scheme = "https"


class Client(object):

    def __init__(self, uri, transports=None, serializer=None):
        if transports is None:
            transports = [HTTPTransport(), HTTPSTransport()]

        if serializer is None:
            serializer = Serializer()

        self._transports = dict([(t.scheme, t) for t in transports])

        parsed = urllib_parse.urlparse(uri)

        if parsed.scheme not in self._transports:
            raise ValueError("Invalid uri scheme {scheme}. Must be one of {available}.".format(scheme=parsed.scheme, available=",".join(self._transports)))

        self._transport = self._transports[parsed.scheme]

        # Default to /RPC2 for path as it is a common endpoint
        if not parsed.path:
            parsed = parsed[:2] + ("/RPC2",) + parsed[3:]

        self._uri = urllib_parse.urlunparse(parsed)

        self._serializer = serializer

    def __call__(self, method, *args):
        data = {
            "methodCall": {
                "methodName": method,
                "params": args,
            }
        }

        body = self._serializer.serialize(data)
        resp = self._transport.request(self._uri, body)
        values = self._serializer.deserialize(resp)

        if "methodResponse" in values:
            values = values["methodResponse"]
        else:
            raise TypeError("Unknown return from xmlrpc call")

        if "params" in values:
            if len(values["params"]) == 1:
                return values["params"][0]
            elif len(values["params"]) > 1:
                return values["params"]

        if "fault" in values:
            raise Fault(values["fault"]["faultString"], code=values["fault"]["faultCode"])

    def __getattr__(self, name):
        return functools.partial(self, name)
